(******************************************************************************
 * Copyright (c) 2017, Aoto Laboratory, Niigata University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the 
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(******************************************************************************
 * file: rwtools/term_rewriting/ctrs.sml
 * description: conditional term rewriting systems
 * author: AOTO Takahito
 * 
 ******************************************************************************)

signature CTRS = 
   sig
   val runDebug: bool ref
   type crule = Term.term * Term.term * (Term.term * Term.term) list
   type crules = crule list

   type ceq = Term.term * Term.term * (Term.term * Term.term) list
   type ceqs = ceq list

   val prRule: crule -> string
   val prRules: crules -> string

   val prEq: ceq -> string
   val prEqs: ceqs -> string

   val condLinearlizationRule: Trs.rule -> crule
   val condLinearlizationRules: Trs.rules -> crules
   val renameRules: crules -> crules

end;

structure Ctrs : CTRS = 
struct

type crule = Term.term * Term.term * (Term.term * Term.term) list
type crules = crule list

type ceq = Term.term * Term.term * (Term.term * Term.term) list
type ceqs = ceq list

val runDebug = ref true : bool ref
fun debug f = if !runDebug then f () else ()
exception CtrsError


local 
    structure T = Term
    structure S = Subst
    structure L = List
    structure LU = ListUtil
    structure LP = ListPair
    structure VM = VarMap
    open PrintUtil
in

fun prCond (l,r) = (T.toString l) ^ " == " ^ (T.toString r)
fun prConds cs = LU.toStringComma prCond cs

fun prRule (l,r,[]) = (T.toString l) ^ " -> " ^ (T.toString r)
  | prRule (l,r,cs) = (T.toString l) ^ " -> " ^ (T.toString r) ^ " | " ^ (prConds cs)
									     
fun prRules rs = LU.toStringCommaLnSquare prRule rs

fun prEq (l,r,[]) = (T.toString l) ^ " = " ^ (T.toString r)
  | prEq (l,r,cs) = (T.toString l) ^ " = " ^ (T.toString r) ^ " | " ^ (prConds cs)
									     
fun prEqs rs = LU.toStringCommaLnSquare prEq rs

fun condLinearlizationRule (l,r) =
    let val lvars = T.varListInTerm l
	val linLhs = T.linearize l
	val freshLinLhs = Subst.renameTermDisjointFrom lvars linLhs
	val renaming = valOf (Subst.match freshLinLhs l)
	(* val cancel = [] *)
        (*** refine the definition of linerization  ***)
	val cancel = LU.differenceByAll' Var.equal (T.linearVarListInTerm l, T.varListInTerm r)
	val renamingCancel = VM.filteri (fn (v,t) => LU.member' Var.equal (valOf (T.varRootOfTerm t)) cancel) renaming
	val renamingRest = VM.filteri (fn (v,t) => not (LU.member' Var.equal (valOf (T.varRootOfTerm t)) cancel)) renaming
	val ans = Subst.applySubst renamingCancel freshLinLhs
	val cond = L.map (fn (x,t) => (T.mkVarTerm (x, Sort.null), t)) (VM.listItemsi renamingRest)
    in (ans, r, cond) 
    end

fun condLinearlizationRules rs = L.map condLinearlizationRule rs

fun renameRules crules = 
    let val f = Fun.fromString "Dummy"
	fun trans (l,r,cs) = let val (csL,csR) = LP.unzip cs
				 val t1 = T.mkFunTerm (f, [l,r],Sort.null)
				 val t2 = T.mkFunTerm (f, csL,Sort.null)
				 val t3 = T.mkFunTerm (f, csR,Sort.null)
			     in T.mkFunTerm (f, [t1,t2,t3], Sort.null)
			     end
	val ts = Subst.renameTerms (L.map trans crules)
	fun back t = let val [t1,t2,t3] = T.argsOfTerm t
			 val [l,r] = T.argsOfTerm t1
			 val csL = T.argsOfTerm t2
			 val csR = T.argsOfTerm t3
		     in (l, r, LP.zip (csL,csR)) 
		     end
    in L.map back ts
    end


end (* of local *)

end; (* of structure Trs *)

